package org.jboss.jms.client;

import javax.jms.*;
import javax.naming.Context;
import javax.naming.InitialContext;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

/**
 * Simple receiver with trans acknowledge session. ABLE to failover.
 *
 * @author mnovak
 */
public class ReceiverTransAck extends Thread {

    private int maxRetries;
    private String hostname;
    private int port;
    private String queueNameJndi = "jms/queue/testQueue1";
    private long receiveTimeOut;
    private int commitAfter;
    private FinalTestMessageVerifier messageVerifier;
    private List<Message> listOfReceivedMessages = new ArrayList<Message>();;
    private List<Message> listOfReceivedMessagesToBeCommited = new ArrayList<Message>();
    private Exception exception = null;
    private long delay = 0;

    private int counter = 0;

    /**
     * Creates a receiver to queue with client acknowledge.
     * 
     * @param hostname hostname
     * @param port jndi port
     * @param queueJndiName jndi name of the queue    
     */
    public ReceiverTransAck(String hostname, int port, String queueJndiName) {
        
        this(hostname, port, queueJndiName, 1000, 10, 5);
        
    }
    
    /**
     * Creates a receiver to queue with trans acknowledge.
     * 
     * @param hostname hostname
     * @param port jndi port
     * @param queueJndiName jndi name of the queue
     * @param receiveTimeOut how long to wait to receive message
     * @param commitAfter send ack after how many messages
     * @param maxRetries how many times to retry receive before giving up
     */
    public ReceiverTransAck(String hostname, int port, String queueJndiName, long receiveTimeOut,
            int commitAfter, int maxRetries) {
        
        this.hostname = hostname;
        this.port = port;
        this.queueNameJndi = queueJndiName;
        this.receiveTimeOut = receiveTimeOut;
        this.commitAfter = commitAfter;
        this.maxRetries = maxRetries;
        
    }

    @Override
    public void run() {

        Context context = null;
        ConnectionFactory cf = null;
        Connection conn = null;
        Session session = null;
        Queue queue = null;

        try {

            final Properties env = new Properties();
            env.put(Context.INITIAL_CONTEXT_FACTORY, "org.jboss.naming.remote.client.InitialContextFactory");
            env.put(Context.PROVIDER_URL, "remote://" + hostname + ":" + port);
            context = new InitialContext(env);

            cf = (ConnectionFactory) context.lookup("jms/RemoteConnectionFactory");

            conn = cf.createConnection();

            conn.start();

            queue = (Queue) context.lookup(queueNameJndi);

            session = conn.createSession(true, Session.SESSION_TRANSACTED);

            MessageConsumer receiver = session.createConsumer(queue);

            Message message = null;
            
            while ((message = receiveMessage(receiver)) != null) {
                
                listOfReceivedMessagesToBeCommited.add(message);
                
                counter++;
                
                Thread.sleep(getDelay());
                
                if (counter % commitAfter == 0) { // try to ack message
                    
                    commitSession(session);
                    
                } else { // count don't want to ack now
                    
                    System.out.println("Receiver for node: " + hostname + " and queue: " + queueNameJndi 
                            + ". Received message - count: "
                            + counter + ", messageId:" + message.getJMSMessageID());
                }
            }
            
            commitSession(session);

            System.out.println("Receiver for node: " + hostname + " and queue: " + queueNameJndi 
                    +". Received NULL - number of received messages: " + counter);
            
            if (messageVerifier != null)    {
                messageVerifier.addReceivedMessages(listOfReceivedMessages);
            }
            
        } catch (JMSException ex) {
            System.out.println("JMSException was thrown during receiving messages:" + ex.getMessage());
            exception = ex;
        } catch (Exception ex) {
            System.out.println("Exception was thrown during receiving messages:" + ex.getMessage());
            exception = ex;
            throw new RuntimeException("Fatal exception was thrown in receiver. Receiver for node: " + hostname);
        } finally {
            if (conn != null) {
                try {
                    conn.close();
                } catch (JMSException ex) {
                    // ignore
                }
            }
            if (context != null)    {
                try {
                    context.close();
                } catch (Exception ex)  {
                    // ignore
                }
            }
        }
    }
    
    /**
     * Try to acknowledge a message.
     * 
     * @throws JMSException
     * 
     */
    public void commitSession(Session session) throws JMSException {
        try {

            session.commit();
            
            System.out.println("Receiver for node: " + hostname + ". Received message - count: "
                            + counter  + " SENT COMMIT");
            
            listOfReceivedMessages.addAll(listOfReceivedMessagesToBeCommited);

        } catch (Exception ex) {
            ex.printStackTrace();
            System.out.println(" Receiver - COMMIT FAILED - Exception thrown during commit: " + ex.getMessage() +". Receiver for node: " + hostname 
                    + ". Received message - count: " + counter);
            // all uncommited messges will be received again
            counter = counter - commitAfter;
            session.rollback();
            
        } finally {
            listOfReceivedMessagesToBeCommited.clear();
        }
    }
    
    /**
     * Tries to receive message from server in specified timeout. If server crashes
     * then it retries for maxRetries. If even then fails to receive which means that
     * consumer.receiver(timeout) throw JMSException maxRetries's times then throw Exception above.
     * 
     * @param consumer consumer message consumer
     * @return message or null
     * @throws Exception when maxRetries was reached
     * 
     */
    public Message receiveMessage(MessageConsumer consumer) throws Exception {
        
        Message msg = null;
        int numberOfRetries = 0;
        
        // receive message with retry
        while (numberOfRetries < maxRetries)    {
            
            try {
                
                msg = consumer.receive(receiveTimeOut);
                return msg;
                
            } catch (JMSException ex)   {
                numberOfRetries++;
                System.out.println("RETRY receive for host: " + hostname + ", Trying to receive message with count: " + (counter + 1));
            }
        }
       
        throw new Exception("FAILURE - MaxRetry reached for receiver for node: " + hostname);
    }
    
    /**
     * @return the maxRetries
     */
    public int getMaxRetries() {
        return maxRetries;
    }

    /**
     * @param maxRetries the maxRetries to set
     */
    public void setMaxRetries(int maxRetries) {
        this.maxRetries = maxRetries;
    }

    /**
     * @return the hostname
     */
    public String getHostname() {
        return hostname;
    }

    /**
     * @param hostname the hostname to set
     */
    public void setHostname(String hostname) {
        this.hostname = hostname;
    }

    /**
     * @return the port
     */
    public int getPort() {
        return port;
    }

    /**
     * @param port the port to set
     */
    public void setPort(int port) {
        this.port = port;
    }

    /**
     * @return the queueNameJndi
     */
    public String getQueueNameJndi() {
        return queueNameJndi;
    }

    /**
     * @param queueNameJndi the queueNameJndi to set
     */
    public void setQueueNameJndi(String queueNameJndi) {
        this.queueNameJndi = queueNameJndi;
    }

    /**
     * @return the messageVerifier
     */
    public FinalTestMessageVerifier getMessageVerifier() {
        return messageVerifier;
    }

    /**
     * @param messageVerifier the messageVerifier to set
     */
    public void setMessageVerifier(FinalTestMessageVerifier messageVerifier) {
        this.messageVerifier = messageVerifier;
    }

    /**
     * @return the listOfReceivedMessages
     */
    public List<Message> getListOfReceivedMessages() {
        return listOfReceivedMessages;
    }

    /**
     * @param listOfReceivedMessages the listOfReceivedMessages to set
     */
    public void setListOfReceivedMessages(List<Message> listOfReceivedMessages) {
        this.listOfReceivedMessages = listOfReceivedMessages;
    }

    /**
     * @return the exception
     */
    public Exception getException() {
        return exception;
    }

    /**
     * @param exception the exception to set
     */
    public void setException(Exception exception) {
        this.exception = exception;
    }

    
    public static void main(String[] args) throws InterruptedException  {

        System.out.println("Parameters are: hostname jndi_name_queue after_how_many_messages_to_commit");
        ReceiverTransAck receiver = new ReceiverTransAck(args[0], 4447, args[1], 10000, Integer.valueOf(args[2]), 10);;
        
        receiver.start();
        
        receiver.join();
    }

    /**
     * @return the delay
     */
    public long getDelay() {
        return delay;
    }

    /**
     * @param delay the delay to set
     */
    public void setDelay(long delay) {
        this.delay = delay;
    }
}
